module MergeArrays
open System

(*
   This file contains three complete implementations of merge sort for
   an array. One is "bottom-up", another makes the dubious choice to
   convert back and forth to lists under the hood (Do you recognize
   the name of the author of this one?). The third is "top-down",
   based on: https://en.wikipedia.org/wiki/Merge_sort#Top-down_implementation
*)


let rec merge (cmp:'a->'a->bool) (a:'a array) ((i,j,k):int*int*int) ((b,m,e):int*int*int)  (a':'a array) =
        //printfn "merge (%d %d %d) (%d %d %d)" i j k b m e
        if i = m && j = e then (*printfn "both exhausted";*) ()
        elif i = m then
            Array.set a' k a.[j]
            //printf "m:%A at %d won by default, now %A\n" a.[j] j a'
            merge cmp a (i,j+1,k+1) (b,m,e) a'
        elif j = e then
            Array.set a' k a.[i];
            //printf "b:%A at %d won by default, now %A\n" a.[i] i a'
            merge cmp a (i+1,j,k+1) (b,m,e) a'
        elif a.[i] <= a.[j] then
            Array.set a' k a.[i];
            //printf "b:%A at %d won versus m:%A at %d, now:%A\n" a.[i] i a.[j] j a';
            merge cmp a (i+1,j,k+1) (b,m,e) a'
        else
            Array.set a' k a.[j];
            //printf "m:%A at %d won versus b:%A at %d, now:%A\n" a.[j] j a.[i] i a';
            merge cmp a (i,j+1,k+1) (b,m,e) a'


let mergesort_bottomup (cmp:'a->'a->bool) (a:'a array) =
    let rec msaux (cmp:'a->'a->bool) (a:'a array) (b:'a array) (n:int) (w:int) : unit =
        let def_generator k =
            let i = k*2*w in
            let lo = i in
            let md = min (i+w) n in
            let hi = min (i+2*w) n in
                merge cmp a (lo, md, lo) (lo, md, hi) b
        in
        
        if w > n || w <= 0 || n <= 0 then
            ()
        else
        (
            let num_merges = if n%(2*w) = 0 then n/(2*w) else n/(2*w)+1  in 
            let tasklist = List.init num_merges def_generator in
            msaux cmp b a n (2*w)
        )
    in
    let n = Array.length a in
    if n <= 1 then ()
    else (
        let a' = Array.copy a in
        // this is silly, because we swap a and b to avoid a copy,
        // we have to start out with a and b in opposite positions
        // depending on how many times we're going to flip them.
        if (int (ceil ( Math.Log (float n))) % 2 = 0) then msaux cmp a' a n 1
        else msaux cmp a a' n 1 

    
    )

// by Xavier -- https://searchcode.com/codesearch/view/961309/
let mergesort_withlists (cmp) (gc:'a array) =
  let rec merge (i:int) (l1:'a list) (l2:'a list) : unit =
    match (l1, l2) with
      ([], []) ->
        ()
    | ([], v2::r2) ->
        Array.set gc i v2; merge (i+1) l1 r2
    | (v1::r1, []) ->
        Array.set gc i v1; merge (i+1) r1 l2
    | (v1::r1, v2::r2) ->
        if cmp v1 v2
        then Array.set gc i v1; merge (i+1) r1 l2 
        else Array.set gc i v2; merge (i+1) l1 r2
  in
  let rec msort start len =
    if len < 2 then () else begin
      let m = len / 2 in
      msort start m;
      msort (start+m) (len-m);
      merge start
            (Array.toList (Array.sub gc start m))
            (Array.toList (Array.sub gc (start+m) (len-m)))
    end in 
  msort 0 (Array.length gc)

let mergesort_topdown (cmp:'a->'a->bool) (a:'a array) =
    let rec msaux cmp a' lo hi a =
        if (hi - lo) < 2 then ()
        else
            let m = (hi + lo) / 2 in
            msaux cmp a lo m a';
            msaux cmp a m hi a';
            merge cmp a' (lo,m,lo) (lo,m,hi) a
    in
    let n = Array.length a in
    if n = 1 then ()
    else
        let a' = Array.copy a in
        msaux cmp a' 0 n a


  
let mergesort = mergesort_topdown
let int_sort : int array -> unit = mergesort (<=)
//let int_data : int array = [|5; 7; 4; 3; 2; 1; 6|]
let int_data : int array =
    let rnd = System.Random() in
    Array.init (int (2.**20.)) (fun _ -> (rnd.Next ()))
let int_sorted : int array = Array.sort (Array.copy int_data)

    
(* returns true if x <= y, where false < true *)
let bool_cmp (x:bool) (y:bool) : bool =
    (not x) || y
  
let bool_sort : bool array -> unit = mergesort bool_cmp
let bool_data : bool array = [|true; true; false; false; true; false|]
let bool_sorted : bool array = Array.sort (Array.copy bool_data)

let test1 () = bool_sort bool_data; bool_data = bool_sorted
let test2 () = int_sort int_data; (*printfn "mine: %A" int_data; printfn "true: %A" int_sorted;*) int_data = int_sorted

let tests = [test1; test2]

let rec runtests (tests : (unit -> bool) list) =
  match tests with
      [] -> ()
    | hd::tail -> 
      if hd () 
      then printfn "success!" 
      else printfn "failure :-(";
      runtests tail


let main () =
    printfn "\n\nBeginning Tests\n\n"; 
    runtests tests;
    printfn "\nCompleted Tests\n\n"

(*
   uncomment the line below to facilitate running this file in isolation, e.g.
   # fsharpc MergeArrays.fs
   # mono MergeArrays.exe
*)
// main ()
